@@ -72,9 +72,9 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
     with open(inFile, 'r') as file:
         inPath = os.path.dirname(os.path.abspath(inFile))
         content = file.read()
-        packages = re.findall('"package://(.*)"', content)
-        if packages:
-            packageName = packages[0].split('/')[0]
+
+        for match in re.finditer('"package://(.*)"', content):
+            packageName = match.group(1).split('/')[0]
             directory = inPath
             while packageName != os.path.split(directory)[1] and os.path.split(directory)[1]:
                 directory = os.path.dirname(directory)
@@ -89,7 +89,7 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
                                      % packageName)
             if os.path.split(directory)[1]:
                 packagePath = os.path.split(directory)[0]
-                content = content.replace('package:/', packagePath)
+                content = content.replace('package://'+packageName, packagePath+'/'+packageName)
             else:
                 sys.stderr.write('Can\'t determine package root path.\n')
 
@@ -149,25 +149,24 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
                     linkList.append(urdf2webots.parserURDF.getLink(link, inPath))
                 for link in linkList:
                     if urdf2webots.parserURDF.isRootLink(link.name, childList):
+                        # We want to skip links between the robot and the static environment.
                         rootLink = link
-                        # if root link has only one joint which type is fixed,
-                        # it should not be part of the model (link between robot and static environment)
-                        while True:
-                            directJoint = []
-                            found = False  # To avoid endless loop
+                        previousRootLink = link
+                        while rootLink in ['base_link', 'base_footprint']:
+                            directJoints = []
                             for joint in jointList:
                                 if joint.parent == rootLink.name:
-                                    directJoint.append(joint)
-                            if len(directJoint) == 1 and directJoint[0].type == 'fixed':
+                                    directJoints.append(joint)
+                            if len(directJoints) == 1:
                                 for childLink in linkList:
-                                    if childLink.name == directJoint[0].child:
+                                    if childLink.name == directJoints[0].child:
+                                        previousRootLink = rootLink
                                         rootLink = childLink
-                                        found = True
-                                        break
                             else:
+                                rootLink = previousRootLink
                                 break
-                            if not found:
-                                break
+
+
                         print('Root link: ' + rootLink.name)
                         break
 
@@ -176,6 +175,7 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
                         urdf2webots.parserURDF.parseGazeboElement(child, rootLink.name, linkList)
 
                 sensorList = (urdf2webots.parserURDF.IMU.list +
+                              urdf2webots.parserURDF.P3D.list +
                               urdf2webots.parserURDF.Camera.list +
                               urdf2webots.parserURDF.Lidar.list)
                 print('There are %d links, %d joints and %d sensors' % (len(linkList), len(jointList), len(sensorList)))
