@@ -1,4 +1,5 @@
 """Import modules."""
+import math
 import os
 import sys
 import struct
@@ -253,23 +254,78 @@ class IMU():
     def export(self, file, indentationLevel):
         """Export this IMU."""
         indent = '  '
+
+        # export InertialUnit
+        file.write(indentationLevel * indent + 'InertialUnit {\n')
+        file.write(indentationLevel * indent + '  name "%s inertial"\n' % self.name)
+        if self.gaussianNoise > 0:
+            file.write(indentationLevel * indent + '  noise %lf\n' % (self.gaussianNoise / (math.pi/2)))
+        file.write(indentationLevel * indent + '}\n')
+
+        # export Accelerometer
         file.write(indentationLevel * indent + 'Accelerometer {\n')
         file.write(indentationLevel * indent + '  name "%s accelerometer"\n' % self.name)
         if self.gaussianNoise > 0:
             file.write(indentationLevel * indent + '  lookupTable [-100 -100 %lf, 100 100 %lf]\n' %
                        (-self.gaussianNoise / 100.0, self.gaussianNoise / 100.0))
         file.write(indentationLevel * indent + '}\n')
+
+        # export Gyro
         file.write(indentationLevel * indent + 'Gyro {\n')
         file.write(indentationLevel * indent + '  name "%s gyro"\n' % self.name)
         if self.gaussianNoise > 0:
             file.write(indentationLevel * indent + '  lookupTable [-100 -100 %lf, 100 100 %lf]\n' %
                        (-self.gaussianNoise / 100.0, self.gaussianNoise / 100.0))
         file.write(indentationLevel * indent + '}\n')
+
+        # export Compass
         file.write(indentationLevel * indent + 'Compass {\n')
         file.write(indentationLevel * indent + '  name "%s compass"\n' % self.name)
         if self.gaussianNoise > 0:
             file.write(indentationLevel * indent + '  lookupTable [-1 -1 %lf, 1 1 %lf]\n' %
-                       -self.gaussianNoise, self.gaussianNoise)
+                       (-self.gaussianNoise, self.gaussianNoise))
+        file.write(indentationLevel * indent + '}\n')
+
+
+class P3D():
+    """Define P3D (ground truth pose)."""
+
+    list = []
+
+    def __init__(self):
+        """Initializatization."""
+        self.name = 'p3d'
+        self.gaussianNoise = 0
+        self.noiseCorrelation = 0
+        self.speedNoise = 0
+        self.parentLink = None
+
+    def export(self, file, indentationLevel):
+        """Export this P3D."""
+        indent = '  '
+
+        # export GPS
+        file.write(indentationLevel * indent + 'GPS {\n')
+        file.write(indentationLevel * indent + '  name "%s gps"\n' % self.name)
+        if self.noiseCorrelation > 0:
+            file.write(indentationLevel * indent + '  noiseCorrelation %lf\n' % self.noiseCorrelation)
+        if self.speedNoise > 0:
+            file.write(indentationLevel * indent + '  speedNoise %lf\n' % self.speedNoise)
+        file.write(indentationLevel * indent + '}\n')
+
+        # export InertialUnit
+        file.write(indentationLevel * indent + 'InertialUnit {\n')
+        file.write(indentationLevel * indent + '  name "%s inertial"\n' % self.name)
+        if self.gaussianNoise > 0:
+            file.write(indentationLevel * indent + '  noise %lf\n' % (self.gaussianNoise / (math.pi/2)))
+        file.write(indentationLevel * indent + '}\n')
+
+        # export Gyro
+        file.write(indentationLevel * indent + 'Gyro {\n')
+        file.write(indentationLevel * indent + '  name "%s gyro"\n' % self.name)
+        if self.gaussianNoise > 0:
+            file.write(indentationLevel * indent + '  lookupTable [-100 -100 %lf, 100 100 %lf]\n' %
+                       (-self.gaussianNoise / 100.0, self.gaussianNoise / 100.0))
         file.write(indentationLevel * indent + '}\n')
 
 
@@ -385,19 +441,11 @@ def getSTLMesh(filename, node):
     stlFile = open(filename, 'rb')
     stlFile.read(80)
     numTriangles = struct.unpack("@i", stlFile.read(4))[0]
-    struct.unpack("<3f", stlFile.read(12))
-    a = struct.unpack("<3f", stlFile.read(12))
-    b = struct.unpack("<3f", stlFile.read(12))
-    c = struct.unpack("<3f", stlFile.read(12))
-    struct.unpack("H", stlFile.read(2))
-    trimesh = node.geometry.trimesh
-    trimesh.coord.append(a)
-    trimesh.coord.append(b)
-    trimesh.coord.append(c)
     if numTriangles > 50000 and not disableMeshOptimization:
         print('Warning: This mesh has a lot of triangles!')
         print('Warning: It is recommended to use the script with the \'--disable-mesh-optimization\' argument.')
-    for i in range(1, numTriangles):
+    trimesh = node.geometry.trimesh
+    for i in range(numTriangles):
         if i % 100 == 0:
             sys.stdout.write('%d / %d\r' % (i, numTriangles))
         struct.unpack("<3f", stlFile.read(12))
@@ -580,6 +628,9 @@ def getColladaMesh(filename, node, link):
                                             except IOError:
                                                 visual.material.texture = ""
                                                 print('failed to open ' + os.path.join(dirname, file))
+                            else:
+                                visual.material.diffuse = colorVector2Instance([1.0, 1.0, 1.0, 1.0])
+
                 link.visual.append(visual)
     else:
         for geometry in list(colladaMesh.scene.objects('geometry')):
@@ -944,8 +995,17 @@ def parseGazeboElement(element, parentLink, linkList):
                     if link.name == name:
                         link.forceSensor = True
                         break
+        elif plugin.hasAttribute('filename') and plugin.getAttribute('filename').startswith('libgazebo_ros_p3d'):
+            p3d = P3D()
+            p3d.parentLink = parentLink
+            if hasElement(plugin, 'topicName'):
+                p3d.name = plugin.getElementsByTagName('topicName')[0].firstChild.nodeValue
+            if hasElement(plugin, "xyzOffsets"):
+                print('\033[1;33mWarning: URDF parser cannot handle \"xyzOffsets\" from p3d!\033[0m')
+            if hasElement(plugin, "rpyOffsets"):
+                print('\033[1;33mWarning: URDF parser cannot handle \"rpyOffsets\" from p3d!\033[0m')
+            P3D.list.append(p3d)
     for sensorElement in element.getElementsByTagName('sensor'):
-        sensorElement = element.getElementsByTagName('sensor')[0]
         if sensorElement.getAttribute('type') == 'camera':
             camera = Camera()
             camera.parentLink = parentLink
@@ -969,7 +1029,7 @@ def parseGazeboElement(element, parentLink, linkList):
                 if hasElement(noiseElement, 'stddev'):
                     camera.noise = float(noiseElement.getElementsByTagName('stddev')[0].firstChild.nodeValue)
             Camera.list.append(camera)
-        elif sensorElement.getAttribute('type') == 'ray':
+        elif sensorElement.getAttribute('type') == 'ray' or sensorElement.getAttribute('type') == 'gpu_ray':
             lidar = Lidar()
             lidar.parentLink = parentLink
             lidar.name = sensorElement.getAttribute('name')
@@ -987,13 +1047,13 @@ def parseGazeboElement(element, parentLink, linkList):
                             maxAngle = float(horizontalElement.getElementsByTagName('max_angle')[0].firstChild.nodeValue)
                             lidar.fov = maxAngle - minAngle
                     if hasElement(scanElement, 'vertical'):
-                        horizontalElement = scanElement.getElementsByTagName('horizontal')[0]
-                        if hasElement(horizontalElement, 'samples'):
+                        verticalElement = scanElement.getElementsByTagName('vertical')[0]
+                        if hasElement(verticalElement, 'samples'):
                             lidar.numberOfLayers = \
-                              int(horizontalElement.getElementsByTagName('samples')[0].firstChild.nodeValue)
-                        if hasElement(horizontalElement, 'min_angle') and hasElement(horizontalElement, 'max_angle'):
-                            minAngle = float(horizontalElement.getElementsByTagName('min_angle')[0].firstChild.nodeValue)
-                            maxAngle = float(horizontalElement.getElementsByTagName('max_angle')[0].firstChild.nodeValue)
+                              int(verticalElement.getElementsByTagName('samples')[0].firstChild.nodeValue)
+                        if hasElement(verticalElement, 'min_angle') and hasElement(verticalElement, 'max_angle'):
+                            minAngle = float(verticalElement.getElementsByTagName('min_angle')[0].firstChild.nodeValue)
+                            maxAngle = float(verticalElement.getElementsByTagName('max_angle')[0].firstChild.nodeValue)
                             lidar.verticalFieldOfView = maxAngle - minAngle
                 if hasElement(rayElement, 'range'):
                     rangeElement = rayElement.getElementsByTagName('range')[0]
