@@ -42,7 +42,7 @@ def header(proto, srcFile=None, robotName='', tags=[]):
     """Specify VRML file header."""
     if srcFile:
         header.sourceFile = srcFile
-    proto.write('#VRML_SIM R2021a utf8\n')
+    proto.write('#VRML_SIM R2021b utf8\n')
     proto.write('# license: Apache License 2.0\n')
     proto.write('# license url: http://www.apache.org/licenses/LICENSE-2.0\n')
     if tags:
@@ -372,6 +372,10 @@ def URDFVisual(proto, visualNode, level, normal=False):
             proto.write((shapeLevel + 2) * indent + 'baseColorMap ImageTexture {\n')
             proto.write((shapeLevel + 3) * indent + 'url [ "' + visualNode.material.texture + '" ]\n')
             proto.write((shapeLevel + 2) * indent + '}\n')
+
+            proto.write((shapeLevel + 2) * indent + 'textureTransform TextureTransform {\n')
+            proto.write((shapeLevel + 3) * indent + 'scale 1 -1\n')
+            proto.write((shapeLevel + 2) * indent + '}\n')
         proto.write((shapeLevel + 1) * indent + '}\n')
 
     if visualNode.geometry.box.x != 0:
